@@ -2,11 +2,14 @@ package Mail::Karmasphere::Client;
 
 use strict;
 use warnings;
-use vars qw($VERSION @ISA @EXPORT_OK %EXPORT_TAGS);
+use vars qw($VERSION @ISA @EXPORT_OK %EXPORT_TAGS
+				%QUEUE @QUEUE $QUEUE);
 use Exporter;
 use Data::Dumper;
 use Convert::Bencode qw(bencode bdecode);
 use IO::Socket::INET;
+use Time::HiRes;
+use IO::Select;
 use constant {
 	IDT_IP4_ADDRESS		=> 0,
 	IDT_IP6_ADDRESS		=> 1,
@@ -17,7 +20,7 @@ use constant {
 
 BEGIN {
 	@ISA = qw(Exporter);
-	$VERSION = "1.09";
+	$VERSION = "1.10";
 	@EXPORT_OK = qw(
 					IDT_IP4_ADDRESS IDT_IP6_ADDRESS
 					IDT_DOMAIN_NAME IDT_EMAIL_ADDRESS
@@ -27,6 +30,9 @@ BEGIN {
 		'all' => \@EXPORT_OK,
 		'ALL' => \@EXPORT_OK,
 	);
+	%QUEUE = ();
+	@QUEUE = ();
+	$QUEUE = 100;
 }
 
 # We can't use these until we set up the above variables.
@@ -58,7 +64,7 @@ sub query {
 	return $self->ask(new Mail::Karmasphere::Query(@_));
 }
 
-sub ask {
+sub send {
 	my ($self, $query) = @_;
 
 	die "Not blessed reference: $query"
@@ -68,18 +74,20 @@ sub ask {
 
 	print STDERR Dumper($query) if $self->{Debug};
 
+	my $id = $query->id;
+
 	my $packet = {
-		_	=> $query->id,
+		_	=> $id,
 		i	=> $query->identities,
 	};
 	$packet->{s} = $query->composites if defined $query->composites;
 	$packet->{f} = $query->feeds if defined $query->feeds;
 	$packet->{c} = $query->combiners if defined $query->combiners;
 	$packet->{fl} = $query->flags if defined $query->flags;
-	print STDERR Dumper($packet) if $self->{Debug};
+	# print STDERR Dumper($packet) if $self->{Debug};
 
 	my $data = bencode($packet);
-	print STDERR Dumper($data) if $self->{Debug};
+	print STDERR ">> $data\n" if $self->{Debug};
 
 	if ($self->{Proto} eq 'tcp') {
 		$data = pack("N", length($data)) . $data;
@@ -88,24 +96,85 @@ sub ask {
 	my $socket = $self->{Socket};
 	$socket->send($data)
 					or die "Failed to send to socket: $!";
-	my $response;
+	return $id;
+}
+
+sub _recv_real {
+	my $self = shift;
+
+	my $socket = $self->{Socket};
+
+	my $data;
 	if ($self->{Proto} eq 'tcp') {
 		my $data;
 		$socket->read($data, 4)
 					or die "Failed to receive length from socket: $!";
 		my $length = unpack("N", $data);
-		$socket->read($response, $length)
-					or die "Failed to receive data from socket: $!";
+		$data = '';
+		while ($length > 0) {
+			my $block;
+			my $bytes = $socket->read($block, $length)
+						or die "Failed to receive data from socket: $!";
+			$data .= $block;
+			$length -= $bytes;
+		}
+		print STDERR "<< $data\n" if $self->{Debug};
 	}
 	else {
-		$socket->recv($response, 8192)
+		$socket->recv($data, 8192)
 					or die "Failed to receive from socket: $!";
 	}
-	my $result = bdecode($response);
-	die $result unless ref($result) eq 'HASH';
-	$result->{query} = $query;
+	my $packet = bdecode($data);
+	die $packet unless ref($packet) eq 'HASH';
 
-	return new Mail::Karmasphere::Response($result);
+	my $response = new Mail::Karmasphere::Response($packet);
+	print STDERR Dumper($response) if $self->{Debug};
+	return $response;
+}
+
+sub recv {
+	my ($self, $query, $timeout) = @_;
+
+	my $id = ref($query) ? $query->id : $query;
+	if ($QUEUE{$id}) {
+		@QUEUE = grep { $_ ne $id } @QUEUE;
+		return delete $QUEUE{$id};
+	}
+
+	my $socket = $self->{Socket};
+
+	$timeout = 60 unless defined $timeout;
+	my $finish = time() + $timeout;
+	my $select = new IO::Select();
+	$select->add($socket);
+	while ($timeout > 0) {
+		my @ready = $select->can_read($timeout);
+
+		if (@ready) {
+			my $response = $self->_recv_real();
+			$response->{query} = $query if ref $query;
+			return $response if $response->id eq $id;
+
+			push(@QUEUE, $id);
+			$QUEUE{$id} = $response;
+			if (@QUEUE > $QUEUE) {
+				my $oid = shift @QUEUE;
+				delete $QUEUE{$oid};
+			}
+		}
+
+		$timeout = $finish - time();
+	}
+
+	return undef;
+}
+
+sub ask {
+	my ($self, $query, $timeout) = @_;
+	my $id = $self->send($query);
+	my $response = $self->recv($query, $timeout);
+	# $response->{query} = $query;
+	return $response;
 }
 
 =head1 NAME
@@ -115,17 +184,26 @@ Mail::Karmasphere::Client - Client for Karmasphere Reputation Server
 =head1 SYNOPSIS
 
 	use Mail::Karmasphere::Client qw(:all);
+
 	my $client = new Mail::Karmasphere::Client(
 			PeerAddr	=> '123.45.6.7',
 			PeerPort	=> 8666,
 				);
+
 	my $query = new Mail::Karmasphere::Query();
 	$query->identity('123.45.6.7', IDT_IP4_ADDRESS);
-	$query->combiner('karmasphere.emailchecker');
-	my $response = $client->ask($query);
+	$query->composite('karmasphere.emailchecker');
+	my $response = $client->ask($query, 60);
 	print $response->as_string;
 
-	my $response = $client->query(...);
+	my $id = $client->send($query);
+	my $response = $client->recv($query, 60);
+	my $response = $client->recv($id, 60);
+
+	my $response = $client->query(
+		Identities	=> [ ... ]
+		Composite	=> 'karmasphere.emailchecker',
+			);
 
 =head1 DESCRIPTION
 
@@ -165,10 +243,23 @@ Set to 1 to enable some wire-level debugging.
 
 =over 4
 
-=item $response = $client->ask($query)
+=item $response = $client->ask($query, $timeout)
 
 Returns a L<Mail::Karmasphere::Response> to a
-L<Mail::Karmasphere::Query>.
+L<Mail::Karmasphere::Query>. This is equivalent to
+
+	$client->recv($client->send($query), $timeout)
+
+=item $id = $client->send($query)
+
+Sends a L<Mail::Karmasphere::Query> to the server, and returns the
+id of the query, which may be passed to recv().
+
+=item $response = $client->recv($id, $timeout)
+
+Returns a L<Mail::Karmasphere::Response> to the query with id $id,
+assuming that the query has already been sent using send(). If no
+matching response is read before the timeout, undef is returned.
 
 =item $response = $client->query(...)
 
